/* --COPYRIGHT--,BSD
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

package com.example.simplelinksdkexplorer.oad;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCharacteristic;
import android.content.Context;
import android.net.Uri;
import android.os.AsyncTask;
import android.os.Bundle;
import android.support.v4.app.Fragment;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.ProgressBar;
import android.widget.TextView;

import com.example.simplelinksdkexplorer.R;
import com.example.simplelinksdkexplorer.gatt.CharacteristicChangeListener;
import com.example.simplelinksdkexplorer.gatt.GattCharacteristicWriteCallback;
import com.example.simplelinksdkexplorer.gatt.GattManager;
import com.example.simplelinksdkexplorer.gatt.operations.GattCharacteristicWriteOperation;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.Semaphore;
import java.util.zip.CRC32;

public class OADDownloadFragment extends Fragment {

    private final String OAD_SERVICE_UUID = "f000ffc0-0451-4000-b000-000000000000";
    private final String OAD_IMG_IDENTIFY_CHAR_UUID = "f000ffc1-0451-4000-b000-000000000000";
    private final String OAD_IMG_BLOCK_CHAR_UUID = "f000ffc2-0451-4000-b000-000000000000";
    private final String OAD_EXT_CTL_CHAR_UUID = "f000ffc5-0451-4000-b000-000000000000";

    private static final String BT_DEVICE = "bt_device";
    private static final String GATT_MANAGER = "gatt_manager";
    private static final String SELECTED_FILE = "selected_file";
    private static final String SELECTED_FILE_TYPE = "selected_file_type";
    private static final String FIRMWARE_TARGET = "firmware_target";
    private static final String FIRMWARE_VERSION = "firmware_version";
    private static final String FIRMWARE_ADDRESS = "firmware_address";

    private final int FIRMWARE_EXTENDED_META_LEN = 20;
    private final int FIRMWARE_EXTENDED_META_HEADER_LEN = 12;
    private final byte FIRMWARE_TYPE_SAP = 0x01;
    private final byte FIRMWARE_TYPE_SNP = 0x03;
    private final byte FIRMWARE_STATUS_EXT = 0x02;
    private final int FIRMWARE_EXTENDED_FEATURE_SEGMENT = 1;
    private final int FIRMWARE_SEGMENT_LIST_OFFSET = 16;

    private final int FIRMWARE_OAD_MAX_PAYLOAD_SIZE = 126;
    private final int FIRMWARE_OAD_MAX_PACKET_SIZE = 130;
    private final byte FIRMWARE_OAD_BLK_SIZE_CMD = 0x13;

    private final byte FIRMWARE_CC2650SNP = 0x03;
    private final byte FIRMWARE_MSP432SAP = 0x01;

    private final byte HEX_DATA_CODE = 0x00;
    private final byte HEX_EOF_CODE = 0x01;
    private final byte HEX_EXT_LINN_ADDR_CODE = 0x04;
    private final byte HEX_START_LINM_ADDR_CODE = 0x05;

    private BluetoothDevice mBtdevice;
    private GattManager mGattManager;

    private TextView oad_download_progress;
    private ProgressBar oad_download_progressbar;
    private TextView oad_download_log;
    private Button cancel_button;
    private Button done_button;

    private boolean metaDataRejected;
    private boolean writeCommandFailed;
    private boolean cancelInvoked;
    private boolean done;

    private long curBlockSize = FIRMWARE_OAD_MAX_PAYLOAD_SIZE;
    private long curPacketSize = FIRMWARE_OAD_MAX_PAYLOAD_SIZE;
    private Semaphore commandSem;
    private Semaphore blockWriteSem;
    private Semaphore waitWriteSem;

    private Uri selectedFile;
    private String selectedFileType;
    private byte firmwareTarget;
    private String firmwareVersion;
    private String firmwareAddress;

    private ByteArrayOutputStream metaImage;
    private ByteArrayOutputStream firmwareImage;

    private OnFragmentInteractionListener mListener;

    public OADDownloadFragment() {
        // Required empty public constructor
    }

    public static OADDownloadFragment newInstance(BluetoothDevice btdevice, Uri file, String filetype, byte fwTarget, String fwVersion, String fwAddress) {
        OADDownloadFragment fragment = new OADDownloadFragment();
        Bundle args = new Bundle();
        args.putParcelable(BT_DEVICE, btdevice);
        args.putParcelable(SELECTED_FILE, file);
        args.putString(SELECTED_FILE_TYPE, filetype);
        args.putByte(FIRMWARE_TARGET, fwTarget);
        args.putString(FIRMWARE_VERSION, fwVersion);
        args.putString(FIRMWARE_ADDRESS, fwAddress);
        fragment.setArguments(args);
        return fragment;
    }

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        setRetainInstance(true);

        if (getArguments() != null) {
            mBtdevice = getArguments().getParcelable(BT_DEVICE);
            selectedFile = getArguments().getParcelable(SELECTED_FILE);
            selectedFileType = getArguments().getString(SELECTED_FILE_TYPE);
            firmwareTarget = getArguments().getByte(FIRMWARE_TARGET);
            firmwareVersion = getArguments().getString(FIRMWARE_VERSION);
            firmwareAddress = getArguments().getString(FIRMWARE_ADDRESS);
        }

        if (savedInstanceState!= null) {

        }
        else {
            metaDataRejected = false;
            writeCommandFailed = false;
            cancelInvoked = false;
            done = false;

            blockWriteSem = new Semaphore(0, true);
            commandSem = new Semaphore(0, true);
            waitWriteSem = new Semaphore(0, true);
        }
    }

    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
        // Inflate the layout for this fragment
        return inflater.inflate(R.layout.fragment_oaddownload, container, false);
    }

    @Override
    public void onViewCreated(View view, Bundle savedInstanceState) {
        oad_download_progressbar = (ProgressBar) getActivity().findViewById(R.id.oad_download_progressbar);
        oad_download_progress = (TextView) getActivity().findViewById(R.id.oad_download_progress);
        oad_download_log = (TextView) getActivity().findViewById(R.id.oad_download_log);
        cancel_button = (Button) getActivity().findViewById(R.id.cancel_button);
        done_button = (Button) getActivity().findViewById(R.id.done_button);

        cancel_button.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                Log.d("FirmwareUpdate", "OAD Firmware update cancelled.");
                if (mListener != null)
                {
                    mListener.OADDownloadCompleted(false);
                }
            }
        });

        if (savedInstanceState!= null) {
            if (done) {
                cancel_button.setVisibility(View.GONE);
                done_button.setVisibility(View.VISIBLE);
                done_button.setOnClickListener(new View.OnClickListener() {
                    @Override
                    public void onClick(View view) {
                        cancelInvoked = true;
                        waitWriteSem.release();
                        blockWriteSem.release();
                        commandSem.release();

                        if (mListener != null) {
                            mListener.OADDownloadCompleted(true);
                        }
                    }
                });
            }
        }
        else {
            mGattManager.addCharacteristicChangeListener(
                    UUID.fromString(OAD_IMG_IDENTIFY_CHAR_UUID),
                    new CharacteristicChangeListener() {
                        @Override
                        public void onCharacteristicChanged(String deviceAddress, BluetoothGattCharacteristic characteristic) {
                            Log.d("FirmwareUpdate", "Notify IMG_IDENTIFY - meta data rejected");
                            metaDataRejected = true;
                            blockWriteSem.release();
                        }
                    }
            );

            mGattManager.addCharacteristicChangeListener(
                    UUID.fromString(OAD_IMG_BLOCK_CHAR_UUID),
                    new CharacteristicChangeListener() {
                        @Override
                        public void onCharacteristicChanged(String deviceAddress, BluetoothGattCharacteristic characteristic) {
                            Log.d("FirmwareUpdate", "Notify IMG_BLOCK - new block request");
                            blockWriteSem.release();
                        }
                    }
            );

            mGattManager.addCharacteristicChangeListener(
                    UUID.fromString(OAD_EXT_CTL_CHAR_UUID),
                    new CharacteristicChangeListener() {
                        @Override
                        public void onCharacteristicChanged(String deviceAddress, BluetoothGattCharacteristic characteristic) {
                            byte[] values = characteristic.getValue();
                            if (values[0] == FIRMWARE_OAD_BLK_SIZE_CMD) {
                                Log.d("FirmwareUpdate", "Notify EXT_CTL - Set Block Size");
                                final int responseBlockSize = ((int) (values[2]) << 8) | values[1];

                                Log.d("FirmwareUpdate", "Block size set to " + responseBlockSize + " bytes");
                                printLogMessage("Block size set to " + responseBlockSize + "bytes\n");

                                curBlockSize = responseBlockSize;
                                curPacketSize = responseBlockSize + 4;

                                commandSem.release();
                            }
                        }
                    }
            );

            metaImage = new ByteArrayOutputStream();
            firmwareImage = new ByteArrayOutputStream();

            printLogMessage("Parsing Firmware Image\n");
            new PrepareFirmwareTask().execute();
        }
    }

    @Override
    public void onAttach(Context context) {
        super.onAttach(context);
        if (context instanceof OnFragmentInteractionListener) {
            mListener = (OnFragmentInteractionListener) context;
        } else {
            throw new RuntimeException(context.toString()
                    + " must implement OnFragmentInteractionListener");
        }
    }

    @Override
    public void onDetach() {
        super.onDetach();
        mListener = null;
    }

    public interface OnFragmentInteractionListener {
        void OADDownloadCompleted(boolean completed);
    }


    private void prepareFirmware(Uri uri, String firmwareType, byte firmwareTarget, String fwVersion, String fwAddress) throws IOException {
        InputStream inputStream = getActivity().getContentResolver().openInputStream(uri);
        BufferedReader reader = new BufferedReader(new InputStreamReader(
                inputStream));
        String line;

        int x;
        x = fwVersion.length();
        if (fwVersion.length() < 4) {
            for (int i = 0; i < 8-x; i++)
                fwVersion = "0" + fwVersion;
        }
        x = fwAddress.length();
        if (fwAddress.length() < 8) {
            for (int i = 0; i < 8-x; i++)
                fwAddress = "0" + fwAddress;
        }

        ByteBuffer fwVersionArray = ByteBuffer.wrap(reverseByteArray(hexStringToByteArray(fwVersion)));
        ByteBuffer fwAddressArray = ByteBuffer.wrap(reverseByteArray(hexStringToByteArray(fwAddress)));

        if (firmwareType.equals("txt")) {

            List<ByteArrayOutputStream> byteArrays = new ArrayList<>();
            List<Long> addresses = new ArrayList<>();
            ByteArrayOutputStream curByteArray = null;
            Long curAddress;

            while ((line = reader.readLine()) != null) {
                if (line.length() == 0)
                    continue;

                if (line.contains("@")) {
                    if (curByteArray != null) {
                        byteArrays.add(curByteArray);
                    }

                    curAddress = Long.parseLong(line.replace("@", ""), 16);
                    addresses.add(curAddress);
                    curByteArray = new ByteArrayOutputStream();
                    continue;
                }

                String[] curSplitLine = line.split(" ");
                for (String split : curSplitLine) {
                    if (split.equals("q"))
                        byteArrays.add(curByteArray);
                    else
                        curByteArray.write((byte)(Short.parseShort(split, 16)));
                }
            }

            /* Calculating how long the meta-data will be */
            long metaLen = FIRMWARE_EXTENDED_META_LEN;

            for (int i = 0; i < addresses.size(); i++) {
                metaLen += FIRMWARE_EXTENDED_META_HEADER_LEN;
                metaLen += byteArrays.get(i).size();
            }

            /* Padding for MTU restrictions */
            long padSize = 0;

            while((metaLen % 4) != 0)
            {
                metaLen += 1;
                padSize += 1;
            }

            /* Length in meta data is in words */
            metaLen = metaLen / 4;

            /* Populating the meta data array */
            metaImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(0x00000000).array());         /* Placeholder for CRC */
            metaImage.write(fwVersionArray.array());                                    /* Version */
            metaImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt((int)metaLen).array());       /* Length */
            metaImage.write(fwAddressArray.array());                                    /* Address */
            metaImage.write(ByteBuffer.allocate(1).order(ByteOrder.LITTLE_ENDIAN).put(firmwareTarget).array());        /* Image type */
            metaImage.write(ByteBuffer.allocate(1).order(ByteOrder.LITTLE_ENDIAN).put(FIRMWARE_STATUS_EXT).array());   /* Status */

            /* Generating the OAD_ExtMeta */
            firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt((int)metaLen).array());
            firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(FIRMWARE_EXTENDED_FEATURE_SEGMENT).array());
            firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(FIRMWARE_SEGMENT_LIST_OFFSET).array());
            firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(0xFFFFFFFF).array());

            /* Generating the OAD_ExtSegmentsList */
            firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(addresses.size()).array());

            /* Keeping track of our offsets */
            long curBinaryOffset = 20 + (addresses.size() * FIRMWARE_EXTENDED_META_HEADER_LEN);

            for (int i = 0; i < addresses.size(); i++) {
                /* Offset of this segment */
                firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt((int)curBinaryOffset).array());
                curBinaryOffset += byteArrays.get(i).size();

                /* Address */
                firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(addresses.get(i).intValue()).array());

                /* Length */
                firmwareImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(byteArrays.get(i).size()).array());
            }

            /* Populating each segment's actual data */
            for (int i = 0; i < byteArrays.size(); i++) {
                firmwareImage.write(byteArrays.get(i).toByteArray());
            }

            /* Adding 0xFF pads to the end of the array to make sure we are aligned */
            for (long i = 0; i < padSize; i++) {
                firmwareImage.write((byte)0xFF);
            }

            /* Add CRC padding */
            CRC32 crc32 = new CRC32();
            crc32.update(firmwareImage.toByteArray());
            long crc32Hash = crc32.getValue();

            byte[] tempMetaArray = metaImage.toByteArray();

            tempMetaArray[0] = (byte) (crc32Hash & 0xFF);
            tempMetaArray[1] = (byte) ((crc32Hash >> 8) & 0xFF);
            tempMetaArray[2] = (byte) ((crc32Hash >> 16) & 0xFF);
            tempMetaArray[3] = (byte) ((crc32Hash >> 24) & 0xFF);

            metaImage = new ByteArrayOutputStream();
            metaImage.write(tempMetaArray);
        }
        else if (firmwareType.equals("hex")) {
            /* Keeping track of the current high */
            long curHigh = 0;
            long lastAddr = 0;

            while ((line = reader.readLine()) != null) {
                ByteArrayOutputStream curBytes = new ByteArrayOutputStream();
                String curLine = "";
                if (!line.contains(":"))
                    continue;
                else
                    curLine = line.replaceAll(":", "");

                byte[] curByteArray = hexStringToByteArray(curLine);

                if (curByteArray.length == 0)
                    continue;

                int curByteCount = curByteArray[0] & 0xFF;
                int curAddr = ByteBuffer.wrap(curByteArray).order(ByteOrder.BIG_ENDIAN).getShort(1) & 0x0000FFFF;
                byte curType = curByteArray[3];
                int curByteEnd = curByteCount + 3;

                if (curByteEnd > 4)
                    curBytes.write(Arrays.copyOfRange(curByteArray, 4, curByteEnd+1));
                else
                    curBytes.write(0x00);

                byte curCRC = (curByteArray[curByteArray.length-1]);

                /* Calculating the checksum */
                curByteEnd = curByteArray.length - 2;
                curByteArray = Arrays.copyOfRange(curByteArray, 0, curByteEnd+1);
                byte calculatedCRC = 0;

                for (byte value : curByteArray) {
                    calculatedCRC += value;
                }

                /* 2s complement */
                calculatedCRC = (byte)((((calculatedCRC & 0xFF) ^ 0xFF) + 1) & 0xFF);

                if (calculatedCRC != curCRC) {
                    return;
                }

                /* Data Type */
                if (curType == HEX_DATA_CODE) {
                    long curIndexAddress = ((curHigh << 16) | curAddr) & 0x00000000FFFFFFFFL;

                    while (curIndexAddress != lastAddr) {
                        firmwareImage.write(0xFF);
                        lastAddr = lastAddr + 1;
                    }

                    firmwareImage.write(curBytes.toByteArray());
                    lastAddr = curIndexAddress + curByteCount;
                }
                else if (curType == HEX_EOF_CODE) {
                    Log.d("PrepareFirmware", "end of file reached");
                    break;
                }
                else if (curType == HEX_EXT_LINN_ADDR_CODE) {
                    curHigh = ByteBuffer.wrap(curBytes.toByteArray()).order(ByteOrder.BIG_ENDIAN).getShort(0) & 0x0000FFFF;
                }
            }

            while ((firmwareImage.size() % 4) != 0) {
                firmwareImage.write(0xFF);
            }

            /* Calculate firmwareImage CRC */
            CRC32 crc32 = new CRC32();
            crc32.update(firmwareImage.toByteArray());
            long crc32Hash = crc32.getValue();

            /* Populating the meta data array */
            metaImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt((int)crc32Hash).array());         /* Placeholder for CRC */
            metaImage.write(fwVersionArray.array());                                    /* Version */
            metaImage.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt((firmwareImage.size()/4)).array());       /* Length */
            metaImage.write(fwAddressArray.array());                                    /* Address */
            metaImage.write(ByteBuffer.allocate(1).order(ByteOrder.LITTLE_ENDIAN).put(firmwareTarget).array());        /* Image type */
            metaImage.write(ByteBuffer.allocate(1).order(ByteOrder.LITTLE_ENDIAN).put(FIRMWARE_STATUS_EXT).array());   /* Status */
        }

        inputStream.close();
    }

    class PrepareFirmwareTask extends AsyncTask<Void, Void, String> {

        @Override
        protected void onPreExecute()
        {
            super.onPreExecute();
        }
        @Override
        protected String doInBackground(Void... arg0)
        {
            try {
                prepareFirmware(selectedFile, selectedFileType, firmwareTarget, firmwareVersion, firmwareAddress);
            }
            catch (IOException e) {
                Log.e("OADDownloadFragment", "IOException when calling sendFirmwareToDevice");
            }

            return "";
        }

        @Override
        protected void onPostExecute(String result)
        {
            super.onPostExecute(result);

            if (metaImage.size() > 0 && firmwareImage.size() > 0)
                new SendFirmwareTask().execute();
        }
    }

    public synchronized void sendFirmwareToDevice() throws IOException {
        byte[] blockSizeCommand = {FIRMWARE_OAD_BLK_SIZE_CMD, FIRMWARE_OAD_MAX_PAYLOAD_SIZE >> 8, FIRMWARE_OAD_MAX_PAYLOAD_SIZE & 0xFF};
        mGattManager.queue(new GattCharacteristicWriteOperation(
                mBtdevice,
                UUID.fromString(OAD_SERVICE_UUID),
                UUID.fromString(OAD_EXT_CTL_CHAR_UUID),
                blockSizeCommand
        ));

        try {
            commandSem.acquire();
            Log.d("FirmwareUpdate", "Command Semaphore acquired!");
        }
        catch (InterruptedException e){
            Log.e("FirmwareUpdate", "Interrupted Exception acquiring commandSem");
        }

        if (cancelInvoked) {
            Log.d("FirmwareUpdate", "Cancel Invoked!");
            return;
        }

        long imageSize = ByteBuffer.wrap(metaImage.toByteArray()).order(ByteOrder.LITTLE_ENDIAN).getInt(6);
        final long totalBlocks = (long)Math.ceil(((double)imageSize*4) / ((double)FIRMWARE_OAD_MAX_PAYLOAD_SIZE));

        final String log_crc32 = "  CRC32: " + String.format("0x%08X", ByteBuffer.wrap(metaImage.toByteArray()).order(ByteOrder.LITTLE_ENDIAN).getInt(0));
        final String log_version = "  Version: " + String.format("0x%04X", ByteBuffer.wrap(metaImage.toByteArray()).order(ByteOrder.LITTLE_ENDIAN).getShort(4));
        final String log_length = "  Length: " + imageSize;
        final String log_address = "  Address: " + String.format("0x%08X", ByteBuffer.wrap(metaImage.toByteArray()).order(ByteOrder.LITTLE_ENDIAN).getInt(10));
        final String log_img_type = "  Image Type: " + String.format("0x%02X",metaImage.toByteArray()[14]);
        final String log_status = "  Status: " + String.format("0x%02X",metaImage.toByteArray()[15]);

        printLogMessage("Meta Data Info:\n");
        printLogMessage(log_crc32 + "\n");
        printLogMessage(log_version + "\n");
        printLogMessage(log_length + "\n");
        printLogMessage(log_address + "\n");
        printLogMessage(log_img_type + "\n");
        printLogMessage(log_status + "\n");

        mGattManager.queue(new GattCharacteristicWriteOperation(
                mBtdevice,
                UUID.fromString(OAD_SERVICE_UUID),
                UUID.fromString(OAD_IMG_IDENTIFY_CHAR_UUID),
                metaImage.toByteArray(),
                new GattCharacteristicWriteCallback() {
                    @Override
                    public void call(byte[] characteristic, int status) {
                        if(status == BluetoothGatt.GATT_SUCCESS) {
                            waitWriteSem.release();
                        }
                        else {
                            writeCommandFailed = true;
                            waitWriteSem.release();
                            blockWriteSem.release();
                        }
                    }
                }
        ));

        try {
            waitWriteSem.acquire();
            Log.d("FirmwareUpdate", "Wait Write Semaphore acquired");
        }
        catch (InterruptedException e) {
            Log.e("FirmwareUpdate", "Interrupted Exception acquiring waitWriteSem");
        }

        if (cancelInvoked) {
            Log.d("FirmwareUpdate", "Cancel Invoked!");
            return;
        }

        printLogMessage("Performing firmware update");

        for (long i = 0; i < totalBlocks; i++) {
            if (cancelInvoked) {
                Log.d("FirmwareUpdate", "Cancel Invoked!");
                return;
            }

            try {
                blockWriteSem.acquire();
                Log.d("FirmwareUpdate", "Block Write Semaphore acquired!");
            }
            catch (InterruptedException e) {
                Log.e("FirmwareUpdate", "Interrupted Exception acquiring blockWriteSem");
            }

            if (cancelInvoked) {
                Log.d("FirmwareUpdate", "Cancel Invoked!");
                return;
            }

            if (metaDataRejected) {
                printLogMessage("ERROR: Meta-data rejected! Aborting update.\n");
                Log.d("FirmwareUpdate", "Metadata rejected!");
                return;
            }

            if (writeCommandFailed) {
                printLogMessage("ERROR: Write command failed!\n");
                Log.d("FirmwareUpdate", "Write command failed!");
                return;
            }

            Log.d("FirmwareUpdate", "Sending Packet " + i);

            long curStartIndex = i * curBlockSize;
            long curEndIndex;

            if ((curStartIndex + curBlockSize) < firmwareImage.size()) {
                curEndIndex = curStartIndex + curBlockSize;
            } else {
                curEndIndex = firmwareImage.size();
            }

            /* Constructing the packet */
            ByteArrayOutputStream curSendPacket = new ByteArrayOutputStream();
            curSendPacket.write(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt((int) i).array());
            curSendPacket.write(Arrays.copyOfRange(firmwareImage.toByteArray(), (int) curStartIndex, (int) curEndIndex));

            while (curSendPacket.size() < curPacketSize) {
                curSendPacket.write((byte)0xFF);
            }

            final long progress = ((i+1) * 100) / totalBlocks;

            if (getActivity() != null) {
                getActivity().runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        oad_download_progressbar.setProgress((int) progress);
                        oad_download_progress.setText(progress + "%");
                    }
                });
            }

            /* Sending the packet */
            mGattManager.queue(new GattCharacteristicWriteOperation(
                    mBtdevice,
                    UUID.fromString(OAD_SERVICE_UUID),
                    UUID.fromString(OAD_IMG_BLOCK_CHAR_UUID),
                    curSendPacket.toByteArray()
            ));
        }
    }

    class SendFirmwareTask extends AsyncTask<Void, Void, String> {

        @Override
        protected void onPreExecute()
        {
            super.onPreExecute();
        }
        @Override
        protected String doInBackground(Void... arg0)
        {
            try {
                sendFirmwareToDevice();
            }
            catch (IOException e) {
                Log.e("OADDownloadFragment", "IOException when calling sendFirmwareToDevice");
            }

            return "";
        }

        @Override
        protected void onPostExecute(String result)
        {
            super.onPostExecute(result);

            done = true;

            if (writeCommandFailed || metaDataRejected) {
                printLogMessage("Firmware update completed with errors!");
            }
            else {
                printLogMessage("Firmware update completed successfully!");
            }

            cancel_button.setVisibility(View.GONE);
            done_button.setVisibility(View.VISIBLE);
            done_button.setOnClickListener(new View.OnClickListener() {
                @Override
                public void onClick(View view) {
                    cancelInvoked = true;
                    waitWriteSem.release();
                    blockWriteSem.release();
                    commandSem.release();

                    if (mListener != null) {
                        mListener.OADDownloadCompleted(true);
                    }
                }
            });
        }
    }


    public void printLogMessage(final String message) {
        final SimpleDateFormat dateFormat = new SimpleDateFormat("HH:mm:ss");

        if (getActivity() != null) {
            getActivity().runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    oad_download_log.append(dateFormat.format(new Date()) + ": " + message);
                }
            });
        }
    }

    public static byte[] hexStringToByteArray(String s) {
        int len = s.length();
        byte[] data = new byte[len/2];
        for (int i = 0; i < len; i += 2) {
            data[i/2] = (byte) ((Character.digit(s.charAt(i), 16) << 4)
                    + Character.digit(s.charAt(i+1), 16));
        }
        return data;
    }

    public byte[] reverseByteArray(byte[] a) {
        for(int i = 0; i < a.length / 2; i++)
        {
            int temp = a[i];
            a[i] = a[a.length - i - 1];
            a[a.length - i - 1] = (byte)temp;
        }

        return a;
    }

    public void setGattManager(GattManager gattManager) {
        mGattManager = gattManager;
    }

}
